#----------------------------------------------------------
#
# Mocap Madness
#
#----------------------------------------------------------
import bpy, os, math, mathutils, time
from bpy.props import *
from mathutils import Vector, Matrix, Euler
from bpy_extras.io_utils import ImportHelper
from xml.etree.ElementTree import Element

import MocapMadness.MMData as mocapmadness

print("MMBVHImport.py........................")


#Load in the "MADNESS"
#mocapmadness.FetchMocapMadness()
#print(test.frog)
#Some helper functions


#
#    class CNode:
#

class CNode:
    def __init__(self, words, parent):
        name = words[1]
        for word in words[2:]:
            name += ' '+word
        
        self.name = name
        self.parent = parent
        self.children = []
        self.head = Vector((0,0,0))
        self.offset = Vector((0,0,0))
        if parent:
            parent.children.append(self)
        self.channels = []
        self.matrix = None
        self.inverse = None
        return

    def __repr__(self):
        return "CNode %s" % (self.name)

    def display(self, pad):
        vec = self.offset
        if vec.length < Epsilon:
            c = '*'
        else:
            c = ' '
        print("%s%s%10s (%8.3f %8.3f %8.3f)" % 
            (c, pad, self.name, vec[0], vec[1], vec[2]))
        for child in self.children:
            child.display(pad+"  ")
        return

    def build(self, amt, orig, parent):
        self.head = orig + self.offset
        if not self.children:
            return self.head
        
        zero = (self.offset.length < Epsilon)
        eb = amt.edit_bones.new(self.name)        
        if parent:
            eb.parent = parent
        eb.head = self.head
        tails = Vector((0,0,0))
        for child in self.children:
            tails += child.build(amt, self.head, eb)
        n = len(self.children)
        eb.tail = tails/n
        self.matrix = eb.matrix.to_3x3()
        mat = Matrix(self.matrix)
        mat.invert()
        self.inverse = mat
        #print(self.matrix)
        #print(self.inverse)
        if zero:
            return eb.tail
        else:        
            return eb.head

#
#    readBvhFile(context, filepath, rot90, scale):
#

Location = 1
Rotation = 2
Hierarchy = 1
Motion = 2
Frames = 3
Status = ""
Deg2Rad = math.pi/180
Epsilon = 1e-5

#rot_mode = "NATIVE"
prev_Euler = [Euler()]
TPose_Euler = [Euler()]
TPoseLoc = Vector([0,0,0])
TPoseRigLoc = Vector([0,0,0])

def arange(x):
    # make the angle stay within 0 to 2Pi try recursion
    if x > 360 :
        return arange(x-360)
    if x < 0 :
        return arange(x + 360)
    
    return x

def generate_paths(self):
    EXTENSION = "bvh"
    directory, file = os.path.split(self.filepath)

    if file and not self.all_in_directory:
        #test for extension
        if os.path.splitext(file)[1].lstrip('.').lower() != EXTENSION:
            return [], directory
        
        return [self.filepath], directory
        
    if not file or self.all_in_directory:
        imagepaths = []
        files_in_directory = os.listdir(directory)
        #clean files from nonimages
        files_in_directory = [file for file in files_in_directory
                              if os.path.splitext(file)[1].lstrip('.').lower()
                              == EXTENSION]
        #clean from unwanted extensions

        for file in files_in_directory:
            imagepaths.append(os.path.join(directory, file))
            
        print(imagepaths)
        return imagepaths, directory

def readBvhFile(context, filepath, rot90, scale,rot_mode,sync,UseTPose,Discard,origin,TargetRigName,OnlyRot):
    
    print("scale=%.1f"%scale)
    #print(context.active_object.name)
    fileName = os.path.realpath(os.path.expanduser(filepath))
    (shortName, ext) = os.path.splitext(fileName)
    if ext.lower() != ".bvh":
        raise NameError("Not a bvh file: " + fileName)
    bvh_name = bpy.path.display_name_from_filepath(filepath)
    print( "Loading BVH file "+ fileName )
    
    time1 = time.clock()
    level = 0
    nErrors = 0
    scn = context.scene
    fp = open(fileName, "rU")
    print( "Reading skeleton" )
    global status
    status = "Reading Skeleton"
    lineNo = 0
    for line in fp: 
        words= line.split()
        lineNo += 1
        if len(words) == 0:
            continue
        key = words[0].upper()
        if key == 'HIERARCHY':
            status = Hierarchy
        elif key == 'MOTION':
            if level != 0:
                raise NameError("Tokenizer out of kilter %d" % level)    
            amt = bpy.data.armatures.new("BvhAmt")
            rig = bpy.data.objects.new("BVH"+bvh_name, amt)
            scn.objects.link(rig)
            scn.objects.active = rig
            bpy.ops.object.mode_set(mode='EDIT')
            root.build(amt, Vector((0,0,0)), None)
            #root.display('')
            bpy.ops.object.mode_set(mode='OBJECT')
            status = Motion
        elif status == Hierarchy:
            if key == 'ROOT':    
                node =CNode(words, None)
                root = node
                nodes = [root]
            elif key == 'JOINT':
                node = CNode(words, node)
                nodes.append(node)
            elif key == 'OFFSET':
                (x,y,z) = (float(words[1]), float(words[2]), float(words[3]))
                if rot90:                    
                    node.offset = scale*Vector((x,-z,y))
                else:
                    node.offset = scale*Vector((x,y,z))
            elif key == 'END':
                node = CNode(words, node)
            elif key == 'CHANNELS':
                oldmode = None
                for word in words[2:]:
                    if rot90:
                        (index, mode, sign) = channelZup(word)
                    else:
                        (index, mode, sign) = channelYup(word)
                    if mode != oldmode:
                        indices = []
                        node.channels.append((mode, indices))
                        oldmode = mode
                    indices.append((index, sign))
            elif key == '{':
                level += 1
            elif key == '}':
                level -= 1
                node = node.parent
            else:
                raise NameError("Did not expect %s" % words[0])
        elif status == Motion:
            if key == 'FRAMES:':
                nFrames = int(words[1])
            elif key == 'FRAME' and words[1].upper() == 'TIME:':
                if sync  :
                    frameTime =  float(words[2])*bpy.context.scene.render.fps
                else :
                    frameTime = 1
                status = Frames
                frame = 0
                t = 0
                bpy.ops.object.mode_set(mode='POSE')
                pbones = rig.pose.bones
                #print(indices)
                #print(indices[0][0])
                #if rot_mode != 'QUATERNION':
                global prev_Euler
                global TPose_Euler
                #print(len(nodes))
                prev_Euler = [Euler() for i in range(len(nodes))]
                TPose_Euler = [Euler() for i in range(len(nodes))]
                
                euler_rot_mode = ""
                for j in range(3):
                     
                    euler_rot_mode += indices[j][0]
                if rot_mode == "NATIVE":
                    rot_mode = euler_rot_mode
                print("rotmode = "+rot_mode)
                #print("rotate_mode = "+rotate_mode)
                for pb in pbones:
                    pb.rotation_mode = rot_mode
                if TargetRigName in bpy.data.objects:
                    TargetRig = bpy.data.objects[TargetRigName]
                    for bone in TargetRig.pose.bones:
                        bone.rotation_mode = rot_mode
                else:
                    TargetRig = None
        elif status == Frames:
            #origin = context.scene.MM_Origin
            if sync and frame > 1:
               addFrame(words, t, nodes, pbones, scale,rot_mode,UseTPose,euler_rot_mode,origin,TargetRig=TargetRig,OnlyRot=OnlyRot)
            else:
               addFrame(words, frame, nodes, pbones, scale,rot_mode,UseTPose,euler_rot_mode,origin,TargetRig=TargetRig,OnlyRot=OnlyRot)
            t += frameTime
            frame += 1
            if sync:
                frame = round(frame,3)
            #print(frameTime)

    fp.close()
    time2 = time.clock()
    if UseTPose :
        #Ok decided to use TPose so make hips at location (0,0,0) at frame 0
        #Still in edit mode
        rootbone = rig.pose.bones[0]
        #was moving rest position to 0,0,0.. might be better to move rig and translate the motion so all start at 0,0,0.
        print("Rootbone loc {0}".format(rootbone.location))

        rootbone.location = (0,0,0)
        first_frame = rig.animation_data.action.frame_range[0]
        rootbone.keyframe_insert('location',  frame=first_frame)
        context.scene.frame_set(first_frame)
        bpy.ops.pose.armature_apply()
    print("Bvh file loaded in %.3f s" % (time2-time1))
    #### remove when get enum fixed
    #rot_mode = "NORMAL"
    bpy.ops.object.mode_set(mode='OBJECT')
    action = rig.animation_data.action
    action.name = "MM_"+bvh_name
    action.pose_markers.new("MOCAP_MADNESS:"+bvh_name)
    bpy.ops.object.select_name(name=rig.name)
    
    if False:
        bpy.ops.action.clean(threshold=1)
        
        
        
        
    ##################3
    '''
    
    Starting to Use XML database to keep track of MocapMadness Actions 
    
    '''
    ####################
    

    
    actionsNode = mocapmadness.data.find("Actions")
    actionNode = Element("Action")
    actionNode.set("name",action.name)
    actionsNode.append(actionNode) 
    
    
    
        
    '''
    
    #Very Wishful Thinking
    action.MocapRigName = StringProperty(
        name="MocapRigName", 
        description="Mocap Rig Action Belongs to")
    action["MocapRigName"] = rig.name
    '''
    
    if Discard:
        # Ditch the rig and armature
        context.scene.objects.unlink(rig)
        bpy.data.objects.remove(rig)
        bpy.data.armatures.remove(amt)
    else :
        if origin:
        # origin set use the original height of hip bone to keep on the floor.
           if rot90: # Z-Up
              rig.location.z = TPoseRigLoc.z
           else :  
              rig.location.y = TPoseRigLoc.y
           #print(TPoseLoc)
        
    return("Bvh file loaded in %.3f s" % (time2-time1))

#
#    channelYup(word):
#    channelZup(word):
#

def channelYup(word):
    if word == 'Xrotation':
        return ('X', Rotation, +1)
    elif word == 'Yrotation':
        return ('Y', Rotation, +1)
    elif word == 'Zrotation':
        return ('Z', Rotation, +1)
    elif word == 'Xposition':
        return (0, Location, +1)
    elif word == 'Yposition':
        return (1, Location, +1)
    elif word == 'Zposition':
        return (2, Location, +1)

def channelZup(word):
    if word == 'Xrotation':
        return ('X', Rotation, +1)
    elif word == 'Yrotation':
        return ('Z', Rotation, +1)
    elif word == 'Zrotation':
        return ('Y', Rotation, -1)
    elif word == 'Xposition':
        return (0, Location, +1)
    elif word == 'Yposition':
        return (2, Location, +1)
    elif word == 'Zposition':
        return (1, Location, -1)

#
#    addFrame(words, frame, nodes, pbones, scale):
#

def addFrame(words, frame, nodes, pbones, scale,rot_mode, UseTPose,euler_rot_mode,origin,TargetRig=None,OnlyRot=False):
    m = 0
    i = 0
    global TPoseLoc,TPoseRigLoc
    for node in nodes:
        name = node.name
        Insert = True  # don't keyframe if not in target rig
        if TargetRig:
            if name not in TargetRig.pose.bones:
                Insert = False
       
        #prev_Euler = Euler()
        try:
            pb = pbones[name]
            
        except:
            pb = None
        if pb:
            for (mode, indices) in node.channels:
                if mode == Location  and ( pb == pbones[0] or not OnlyRot) :
                    vec = Vector((0,0,0))

                    for (index, sign) in indices:
                        vec[index] = sign*float(words[m])
                        m += 1
                   
                    location = node.inverse * (scale * vec - node.offset) # FIX for 2.59*  node.inverse
                    #set the first frame as the transition vector
                    if frame < 0.001 :
                       TPoseLoc = location
                       pb.location = location
                       TPoseRigLoc = scale*vec
                       print("setting TPoseRigLoc %s %s"%(vec,TPoseLoc))
                       #TPoseLoc = Vector((0,0,0))
                    pb.location = location
                    if origin:
                       pb.location -= TPoseLoc
                                        
                    
                    if Insert:    
                        pb.keyframe_insert('location',  frame=frame)
                elif mode == Rotation:
                    mats = []

                    for (axis, sign) in indices:
                        angle = arange(sign*float(words[m]))*Deg2Rad
                        
                        mats.append(Matrix.Rotation(angle, 3, axis))
                        m += 1
                    mat = node.inverse * mats[0] * mats[1] * mats[2]* node.matrix
 

                    #print(i)
                    #print(len(prev_Euler))
                    global TPose_Euler
                    if frame > 0.001 and UseTPose == True:
                        #print(TPose_Euler[i].to_3x3())
                        euler = mat.to_euler(euler_rot_mode,prev_Euler[i])
                        diff = TPose_Euler[i]
                        euler = Euler([euler[0]-diff[0],euler[1]-diff[1],euler[2]-diff[2]])
                        euler.order = euler_rot_mode
                        #print(con)
                    else :
                       
                        euler = mat.to_euler(euler_rot_mode,prev_Euler[i])
                    if frame < 0.001 and UseTPose == True:
                        #print(frame)
                        TPose_Euler[i] = euler
                    
                    prev_Euler[i] = euler
                    if Insert:
                        if rot_mode == "QUATERNION":
                            pb.rotation_quaternion = euler.to_quaternion()
                            
                            pb.keyframe_insert('rotation_quaternion',frame=frame)
                            
                        else:
                            pb.rotation_euler = euler
                            pb.keyframe_insert('rotation_euler',frame=frame)
                        #for n in range(3):
                            #pb.keyframe_insert('rotation_euler',index=n,frame=frame)
            i+=1

    return


#
#    class LoadBvhButton(bpy.types.Operator):
#

class LoadBvhButton(bpy.types.Operator, ImportHelper):
    bl_idname = "mocapmadness.import_bvh"
    bl_label = "Load BVH file (.bvh)"
    mocapmadness_name = bl_idname
    for cmd in mocapmadness.SetOperatorProperties("mocapmadness.import_bvh"):
        exec(cmd)
    filename_ext = ".bvh"
    filter_glob = StringProperty(default="*.bvh", options={'HIDDEN'})
    rot_mode = EnumProperty(items=(
        ('QUATERNION', "Quaternion", "Convert rotations to quaternions"),
        ('NATIVE', "Euler (Native)", "Use the rotation order defined in the BVH file"),
        ('XYZ', "Euler (XYZ)", "Convert rotations to euler XYZ"),
        ('XZY', "Euler (XZY)", "Convert rotations to euler XZY"),
        ('YXZ', "Euler (YXZ)", "Convert rotations to euler YXZ"),
        ('YZX', "Euler (YZX)", "Convert rotations to euler YZX"),
        ('ZXY', "Euler (ZXY)", "Convert rotations to euler ZXY"),
        ('ZYX', "Euler (ZYX)", "Convert rotations to euler ZYX"),
        ),
            name="Rotation",
            description="Rotation conversion.",
            default='NATIVE')

    
    all_in_directory = BoolProperty(
            name="Load All In Directory",
            default=False, 
            description="Rotate the armature to make Z point up")
        
    filepath = bpy.props.StringProperty(name="File Path", 
        maxlen=1024, default="")
    '''        
    Scale = bpy.props.FloatProperty(
        name="Scale", 
        default =0.1,
        min = 0.01,
        max = 100)
            
    Rot90 = bpy.props.BoolProperty(
        name="Rotate 90 degrees", 
        default=True,
        description="Rotate the armature to make Z point up")
        
        
    Sync = bpy.props.BoolProperty(
        name="Use timing from bvh",
        default=True, 
        description="Use the frame time from bvh.. otherwise keyframe each frame")              
    Origin = bpy.props.BoolProperty(
        name="Travel Origin", 
        default=True,
        description="Make the starting position (0,0) and original hip height")
    
    
    TPose = bpy.props.BoolProperty(
        name="TPose Rest Pose", 
        default=True,
        description="Attempt to change rest pose to TPose in frame 0")
   
    
    Discard = bpy.props.BoolProperty(
        name="Discard Rig on Import",
        default=False, 
        description="Keep the action but discard rig and armature")

    
    UseOnlyRootBoneLoc = bpy.props.BoolProperty(
        name="Only rootbone locations", 
        default=False,
        description="Only calculate locations for root bones.")
    '''  
    def draw(self, context):
        #print("draw")
        layout = self.layout
        #layout.prop(mocapmadness.blender,"rot_mode")
        box = layout.box()
        box.label('Import Options:', icon='FILTER')
        box.prop(self, 'all_in_directory')
        if self.all_in_directory :
           box.label("With all in directory selected only one rig will be imported, the rest will be imported as Actions Only",icon="INFO")
        
        box.prop(self, 'Dismiss', icon='ACTION')
        box = layout.box()
        box.label('Rig Settings', icon='ARMATURE_DATA')
        box.prop(self,'TargetRig')
        box.prop(self, 'rot_mode', expand=False)
        box.prop(self, 'Scale')
        box.prop(self, 'Rot90',icon="MANIPUL")
        box.prop(self, "TPose",icon="ARMATURE_DATA")
        box = layout.box()
        box.label('Action Settings:', icon='ACTION')
        box.prop(self, 'Discard', icon='ACTION')
        box.prop(self, 'Sync',icon="MANIPUL")
        box.prop(self, 'Origin',icon="MANIPUL")        
        box.prop(self,'UseOnlyRootBoneLoc',icon="MANIPUL")
        

    def execute(self, context):
        import bpy, os
        
        if not "MocapMadness" in context.scene:
            mocapmadness.InitMocapMadness(context.scene)
        
        
        #print(self.properties.rot_mode)
        #context.scene.MM_Scale = self.properties.global_scale
        # If read dir is selected read the entire directory of mocaps and If there is an active rig then only as actions.. If there is no active build rig on first then only actions for rest
        #if self.all_in_directory :
                
        #print(self.files)
        mocapfilelist,path = generate_paths(self)
        first = True
        for mocapfile in mocapfilelist:
           #print(mocapfile) 
                       
           self.report("INFO",readBvhFile(context, mocapfile, self.Rot90,       self.Scale,self.rot_mode,self.Sync,self.TPose,self.Discard,self.Origin,self.TargetRig,self.UseOnlyRootBoneLoc))  
           self.report("INFO","YEAH BABY")
        mocapmadness.Save()
        mocapmadness.PreserveSettings(self)
        if not self.Dismiss:
        
            return bpy.ops.mocapmadness.import_bvh('INVOKE_DEFAULT')
        return{'FINISHED'}    

    def invoke(self, context, event):
        #self.Origin = True  <WORKS CAN SET FROM XML HERE>
        mocapmadness.SetOperatorSettings(self)   
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}   

   
     
#### Register
def MocapMadnessImport(self, context): 
    self.layout.operator(LoadBvhButton.bl_idname, text="Mocap Madness", icon='PLUGIN') 
    
def register():
    #bpy.utils.register_class(BvhImportPanel)
    bpy.utils.register_class(LoadBvhButton)
    bpy.types.INFO_MT_file_import.append(MocapMadnessImport) 
    
def unregister():
    bpy.utils.unregister_class(LoadBvhButton)
    bpy.types.INFO_MT_file_import.remove(MocapMadnessImport)


#register()