import bpy, mathutils
import bpy.ops
from bpy.props import *
from xml.etree.ElementTree import Element
import MocapMadness.MMData as mocapmadness
from math import floor

def resample(action,fps=24):
    # create a new action
    name = action.name
    action.name = name+"_OLD"
    newaction = bpy.data.actions.new(name)
    for fcurve in action.fcurves:
        newfcurve = newaction.fcurves.new(data_path=fcurve.data_path,index=fcurve.array_index)
        #newfcurve.data_path = fcurve.data_path
        #newfcurve.array_index = fcurve.array_index
        i = floor(action.frame_range[0])
        
        while i < action.frame_range[1]:
            bpy.context.scene.frame_set(i)
            newfcurve.keyframe_points.insert(frame=i,value=fcurve.evaluate(i))
            i+=1
    return newaction 
        

class CleanFCurvePanel(bpy.types.Panel):
    bl_label = "Clean Fcurve"
    bl_space_type = "GRAPH_EDITOR"
    bl_region_type = "UI"
    def draw(self, context):
        layout = self.layout


        layout.operator("mocapmadness.curve_cleaner")

    @classmethod 
    def poll(self,context):
        #Fix this to poll for a mocapmadness rig perhaps.
        return True 

class FCurveCleaner(bpy.types.Operator):
    bl_idname = "mocapmadness.curve_cleaner"
    bl_label = "Clean FCurve"
    mocapmadness_name = bl_idname
    if mocapmadness.settings :
        for cmd in mocapmadness.SetOperatorProperties("mocapmadness.curve_cleaner"):
            exec(cmd)
    #threshold = FloatProperty("threshold",default=0.001,description="Threshold for the clean operator",min=0.001,max=0.1,precision=3)
    #self.threshold = 0.001
    
    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self)

    def draw(self,context):
        layout = self.layout
        layout.prop(self,"PreClean")
        layout.prop(self,"pre_threshold")
        layout.prop(self,"PreSmooth")
        layout.prop(self,"Resample")
        layout.prop(self,"PostClean")
        layout.prop(self,"post_threshold")

        print("DRAEWWWW")    
    def execute(self, context):
        if not mocapmadness.data:
            return {'CANCELLED'}
        
        #Quick loop thru the actions in MocapMadnessXML
        
        Actions = mocapmadness.data.findall("Actions/Action")   # list of actions
        rig = bpy.context.object
        #start_action = rig.animation_data.action
        for Action in Actions:
            action = bpy.data.actions[Action.get("name")]
            rig.animation_data.action = action

            if self.PreClean:
                bpy.ops.graph.clean(threshold=self.pre_threshold)
            if self.PreSmooth:
                bpy.ops.graph.smooth()
            if self.Resample:
                print("ACTION RESAMPLED")
                rig.animation_data.action = resample(action)

            if self.PostClean:
                bpy.ops.graph.clean(threshold=self.post_threshold)
            if self.PostSmooth:
                bpy.ops.graph.smooth()            
        
        #rig.animation_data.action = start_action 
        self.report("INFO","%d Actions Cleaned"%len(Actions))
        mocapmadness.PreserveSettings(self)
        return{'FINISHED'}

     
def register():
    bpy.utils.register_class(CleanFCurvePanel)
    bpy.utils.register_class(FCurveCleaner)


def unregister():
    bpy.utils.unregister_class(CleanFCurvePanel)    
    bpy.utils.unregister_class(FCurveCleaner)