# Make a menu from the xml data.

import bpy
from xml.etree.ElementTree import fromstring,Element
from bpy.props import *
from mathutils import Euler,Matrix
import MocapMadness.MMData as mocapmadness
from math import degrees
from MocapMadness.utils import AddHipLocator, SetInterpolation
'''
An 

'''


def selected_to_context(context):
    obj = context.selected_objects[0]
    context.active_object = obj
    
class PlayMocapList(bpy.types.Operator): 
    bl_idname = 'mocapmadness.playmocap' 
    bl_label = 'Play Mocaps'
    createTrack = BoolProperty("CreateTrack",default=False)
    addHipLocator = BoolProperty("AddHipLocator",default=False)
    
    action = StringProperty("Action",default="Select Action") 

        
    def execute(self, context):

       print("Playing Mocaps")
       #get all the actions that have markers named after the rig
       rig = context.selected_objects[0]
       
       if self.createTrack :
           rig = context.selected_objects[0]
           
           RigTrack,BvhTrack,RigAction = MM_get_tracks(rig)
           action = rig.animation_data.action
           rig.animation_data.action = RigAction
           rig.keyframe_insert("location")
           rig.keyframe_insert("rotation_euler",index=2)
           current_frame = context.scene.frame_current
           
           actionStrip = BvhTrack.strips.new("Start",start=current_frame,action=action)
           rigStrip = RigTrack.strips.new("Rig%sStrip"%rig.name,start=current_frame,action=RigAction)
           rig.animation_data.action = None
           RigTrack.strips[0].action_frame_end = actionStrip.frame_end
           return{'FINISHED'}
           
       if self.addHipLocator:
           rig = context.selected_objects[0]
           AddHipLocator(rig,rig.pose.bones[0])
           return{'FINISHED'}    
       
       action = bpy.data.actions[self.action]
       if not action:
           return{'CANCELLED'}
       if not rig.animation_data:
           rig.animation_data_create()
       rig.animation_data.action = action
       context.scene.frame_set(1)
       context.scene.frame_end = action.frame_range[1]
       #bpy.ops.screen.animation_cancel()  
       mode = rig.mode
       bpy.ops.object.mode_set(mode="POSE")
       if mode == "POSE":
           bpy.ops.pose.paths_clear()
           bpy.ops.pose.select_all(action='DESELECT')
    
           rig.pose.bones[0].bone.select = True
           bpy.ops.pose.paths_calculate()
       #bpy.ops.screen.animation_play()
       bpy.ops.object.mode_set(mode=mode)
       #Playing = True
       #for action in bpy.data.actions :
          

       return {'FINISHED'} 
   
def MM_get_tracks(rig):
    ### create the RigTrack and BvhTrack for this rig
    
    print("CREATE TRACK %s"%rig.name)
    if "BvhTrack" in rig.animation_data.nla_tracks:
        BvhTrack = rig.animation_data.nla_tracks["BvhTrack"]
    else:
        BvhTrack = rig.animation_data.nla_tracks.new()
        BvhTrack.name = "BvhTrack"      
  
    if "RigTrack" in rig.animation_data.nla_tracks:
        RigTrack = rig.animation_data.nla_tracks["RigTrack"]
    else:
        RigTrack = rig.animation_data.nla_tracks.new(BvhTrack)
        RigTrack.name = "RigTrack"
    
        
    RigActionName = 'MM_%s_RigAction'%rig.name  
    if RigActionName in bpy.data.actions:
        RigAction = bpy.data.actions[RigActionName]
    else:
        RigAction = bpy.data.actions.new(RigActionName)
        
    return RigTrack,BvhTrack,RigAction
    
def createbranch(strip,branchnode,rig=None,FixGimble=False,FixGimbleAngle=0):
       
    current_frame = bpy.context.scene.frame_current
    rig = bpy.context.selected_objects[0]
    
        
    action_start_frame = float(branchnode.get("from"))
    action2_name = branchnode.get("action")
    
    action2 = bpy.data.actions[action2_name]
    action2_start_frame = float(branchnode.get("to"))
    RigTrack,BvhTrack,RigAction = MM_get_tracks(rig)
    rig.animation_data.action = RigAction

    locator = AddHipLocator(rig,rig.pose.bones[0])
    # current rig location
    
    bpy.context.scene.frame_set(current_frame)    
    locA = locator.matrix_world.to_translation()  # Target destination
    print("ROT:",rig.pose.bones[0].rotation_euler,rig.pose.bones[0].location)
    trans = Matrix.Translation(locA)
    
    rotzA = locator.matrix_world.to_euler("XYZ").z    
    #The current strip Gets taken back one.
    
    strip.action_frame_end = action_start_frame - 0.01
    strip.frame_end = current_frame - 0.01
    branchStrip = BvhTrack.strips.new("Action",start=current_frame,action=action2)
    #branchStrip.frame_end = 
    branchStrip.action_frame_start = action2_start_frame
    branchStrip.action_frame_end = int(action2.frame_range[1])
    #print("BRANCH TO ACTION %s %s) @%.1f"%(action2.name,action2.frame_range,action2_start_frame))
    #branchStrip.action_frame_end = action_end_frame

    bpy.context.scene.frame_set(current_frame)        
    #bpy.context.scene.update()
    # Get the change in rotation
    rotzB = locator.matrix_world.to_euler("XYZ").z
    
    drotz = rotzB - rotzA
    if FixGimble:
        print("FIX GIMBAL:",FixGimbleAngle)
        drotz += FixGimbleAngle
    #rotate it

    
    #get the position the locator is now
             
    bpy.context.scene.frame_set(current_frame)    
    locB = locator.matrix_world.to_translation()

 
 
    rotmat = Matrix.Rotation(drotz,3, "Z")  
    dloc = locB - locA   # translation


    print("BRANCH %s %s"%(dloc,degrees(drotz)))
    #return

    #rig.animation_data.use_nla = False 

    #rig.location -= dloc
    #rig.location = rig.location * rotmat * trans
    rig.rotation_euler.z -= drotz



    rig.keyframe_insert("rotation_euler",index=2,frame=current_frame)
    bpy.context.scene.frame_set(current_frame)
    locC = locator.matrix_world.to_translation()
    dloc = locC - locA
    print(dloc)
    rig.location -= dloc    
    rig.keyframe_insert("location",frame=current_frame)

    print("ROT:",rig.pose.bones[0].rotation_euler,rig.pose.bones[0].location)
    if True:    
        bpy.context.scene.objects.unlink(locator)
        bpy.data.objects.remove(locator)
    

    SetInterpolation(RigAction,"CONSTANT")
    RigTrack.strips[0].action_frame_end = branchStrip.frame_end    
    #rig.animation_data.action = None
    rig.animation_data.use_nla = True

    
def createcycle(strip,cyclenode,repeat=10):
    REPEAT = repeat
    # rig is the mocapmadness rig
    # cyclenode xml cycle node
    # location
    
    action = bpy.data.actions[strip.action.name]
    
    current_frame = bpy.context.scene.frame_current
    rig = bpy.context.selected_objects[0]
    action_start_frame = float(cyclenode.get("from"))
    
    if action_start_frame != strip.strip_time:
        print("ERROR: %.1f %.1f action start times non matching" \
            %(action_start_frame,strip.strip_time))
    
    print("STRIP FRAME %s %.1f"%(strip.name,strip.frame_end))
    action_end_frame = float(cyclenode.get("to"))
    
    print("CYCLE: from %.1f to %.11f"%(action_start_frame,action_end_frame))
    
    start_frame = current_frame + action_start_frame
    end_frame = current_frame + action_end_frame - action_start_frame
    
    locator = AddHipLocator(rig,rig.pose.bones[0])
    #use the hip locator to get the dzrot and translation vector.
    
    RigTrack,BvhTrack,RigAction = MM_get_tracks(rig)
    rig.animation_data.action = RigAction     
    
    

      
    # PUT THE RIG IN THE SLOT
    #rig.animation_data.action = action
    bpy.context.scene.frame_set(current_frame) 
    #bpy.context.scene.frame_set(1.00)
    loc0 = rig.matrix_world.to_translation()
    
    rotzA = locator.matrix_world.to_euler("XYZ").z
    
    
    #bpy.context.scene.frame_set(start_frame)
    locA = locator.matrix_world.to_translation()
    
    
    bpy.context.scene.frame_set(end_frame)
    locB = locator.matrix_world.to_translation()
    print("Loc A %s Loc B %s"%(locA,locB))
    rotzB = locator.matrix_world.to_euler("XYZ").z
    
    drotz = rotzB - rotzA # Z rotation
    
    #drot = Euler([0,0,drotz],"XYZ")

    dloc  = locA - loc0  # translation vector

    #print(drotz)
    #print(dloc*drot.to_matrix())
    #print(dloc*rotmat)

    strip.action_frame_end = action_start_frame
    strip.frame_end = current_frame
    bpy.context.scene.frame_set(current_frame)
    cycleStrip = BvhTrack.strips.new("Cycle",start=current_frame,action=action)
    cycleStrip.action_frame_start = action_start_frame
    cycleStrip.action_frame_end = action_end_frame - 1.0
    cycleStrip.repeat = REPEAT 



    destVec = locB - loc0

    print("Cycle transform vec:%s %s rot:%s"%(dloc,destVec,drotz))
    rotmat = Matrix.Rotation(drotz,3, "Z")
    length = action_end_frame - action_start_frame - 1 
    bpy.context.scene.frame_set(current_frame)
    i = 0
    
    
    # SET UP A FLAG FOR THIS 
    
    if False:
        #take out z loc difference
        destVec[2] = 0
        dloc.z = 0
    #rig.keyframe_insert("location",frame=current_frame+i*length+1)
    
    bpy.context.scene.objects.unlink(locator)
    bpy.data.objects.remove(locator)
    #rig.keyframe_insert("rotation_euler",index=2,frame=current_frame+i*length+1)


    i = 1
    while i < REPEAT:


        dloc_r = (destVec -rotmat* dloc)
        rig.location +=  dloc_r
        rig.rotation_euler.z += drotz
        rig.keyframe_insert("location",frame=current_frame+i*length)
        rig.keyframe_insert("rotation_euler",index=2,frame=current_frame+i*length)
        destVec = rotmat*destVec
        dloc =rotmat* dloc
        i+=1
    SetInterpolation(RigAction,"CONSTANT")
            
    #bpy.ops.action.interpolation_type(type="CONSTANT")
    #rigStrip = RigTrack.strips.new("blah",start=1,action=RigAction)
    
    # TaCK the rEsTof the action onto the end
    frame=current_frame+i*length
    cycleStrip = BvhTrack.strips.new("Cycle",start=frame,action=action)
    cycleStrip.action_frame_start = action_end_frame
    #cycleStrip.action_frame_end = action_end_frame - 1.0
    #cycleStrip.repeat = REPEAT     
    
    RigTrack.strips[0].frame_end = cycleStrip.frame_end
    rig.animation_data.use_nla = True
    #rig.animation_data.action = None        
          
    return False
    

class View3DPanel():
    bl_space_type = 'NLA_EDITOR'
    bl_region_type = 'UI'

    @classmethod
    def poll(cls, context):
        return (context.object is not None)



class NLAPlayMocap(View3DPanel, bpy.types.Panel):
    bl_idname = "VIEW3D_PT_test_1"
    bl_label = "Play Mocap"
    @classmethod
    def poll(self,context):
        if context.selected_objects[0].animation_data:
           return True
       
        return False
    def draw(self, context):
        layout = self.layout
        self.layout.label("MocapMadness Player")
        
        ad = context.selected_objects[0].animation_data or None
        layout.menu("mocapmadness.actionmenu")
        if ad:
            if ad.action:
                layout.prop(ad.action,"name")
            layout.prop(context.selected_objects[0],"name")
        layout.operator("screen.animation_play",text="PLAY")
        layout.operator("mocapmadness.playmocap","HipLocator").addHipLocator=True
        layout.operator("mocapmadness.playmocap","ADD NEW").createTrack=True
        layout.operator("mocapmadness.posecompare")
        layout.operator("mocapmadness.posecompare",text="Clean").clean=True
        #self.layout.operator("mocapmadness.compareposes")

def getCurrentActionFromStrips(context,TrackName,lastStrip=False):
    # look for a rig in the selected_objects[0]
    if not context.selected_objects:
        return False,None
    rig = context.selected_objects[0]
    if rig:
        if rig.animation_data:
            #print("RIG WITH ANIMDATA:%s"%rig.animation_data.nla_tracks)
            #get the active strip
            if TrackName in  rig.animation_data.nla_tracks:
                strips = rig.animation_data.nla_tracks[TrackName].strips
                if lastStrip and len(strips) > 0:
                    strip = strips[len(strips)-1]
                    if strip.action:
                        frame_current = context.scene.frame_current
                        #print("Frame %.1f %s"%(frame_current,strip.action.name))
                        if int(frame_current) in range(int(strip.frame_start),int(strip.frame_end)):
                            return True, strip
                else:
                    for strip in strips:
                        #print(strip.name)
                        if strip.action:
                            frame_current = context.scene.frame_current
                            #print("Frame %.1f %s"%(frame_current,strip.action.name))
                            if int(frame_current) in range(int(strip.frame_start),int(strip.frame_end)):
                                return True, strip
    return False,None
    
    
class CyclePanel(View3DPanel, bpy.types.Panel):
    bl_idname = "VIEW3D_PT_test_2"
    bl_label = "Choose a Cycle"
    @classmethod 
    def poll(self,context):

               
        return getCurrentActionFromStrips(context,"BvhTrack",lastStrip=True)[0]
     
    def draw(self, context):
        #self.layout.label("Cycles")
        layout = self.layout
        
        layout.menu("mocapmadness.cyclemenu")
        
        
       
        
        frame_current = context.scene.frame_current  
        strip = getCurrentActionFromStrips(context,"BvhTrack",lastStrip=True)[1]      
        action = strip.action
        striptime = strip.strip_time
        if action :
            box = layout.box()
            cycles = mocapmadness.data.findall('Actions/Action[@name="%s"]/cycle[@from="%.1f"]'%(action.name,striptime))        
            
            i = 0
            for cycle in cycles:
                
                text = "Cycle To %s (%s)"%(cycle.get("to"),cycle.get("match"))  
                box.operator("mocapmadness.cycle", text=text).cyclenode=i
                #layout.operator("object.select_random", text="Random")
                i += 1
            
            
        
        layout.operator("mocapmadness.cycle",text="Next").next = True     
        
class BranchPanel(View3DPanel, bpy.types.Panel):
    bl_idname = "MocapMadness_BranchPanel"
    bl_label = "Choose a Branch"

    @classmethod 
    def poll(self,context):
        
        return getCurrentActionFromStrips(context,"BvhTrack",lastStrip=True)[0]
    def draw_header(self,context):
        layout = self.layout
        layout.menu("mocapmadness.branchmenu")
        
    def draw(self, context):
        #self.layout.label("Cycles")
        layout = self.layout
        
        layout.menu("mocapmadness.branchmenu")
        
        
        box = layout.box()
        
        #frame_current = context.scene.frame_current 
        
        strip = getCurrentActionFromStrips(context,"BvhTrack",lastStrip=True)[1]      
        action = strip.action 
              
        #action = context.object.animation_data.action
        Branches = mocapmadness.data.findall('Actions/Action[@name="%s"]/branch[@from="%.1f"]'%(action.name,strip.strip_time))        
        
        i = 0
        for Branch in Branches:
            
            text = "Branch To %s:%s (%s)"%(Branch.get("action"),Branch.get("to"),Branch.get("match"))  
            box.operator("mocapmadness.branch", text=text).branchnode=i
            #layout.operator("object.select_random", text="Random")
            i += 1
        
        
        
        layout.operator("mocapmadness.cycle",text="Next").next = True 
                
class ActionMenu(bpy.types.Menu):
    bl_idname = "mocapmadness.actionmenu"
    bl_label = "Change Action"
    filter = 'Actions/Action'
    
    def draw(self,context):
        
        print("FILTER:%s"%ActionMenu.filter)
        print(self.filter)
        
        layout = self.layout

        Actions = mocapmadness.data.findall(self.filter)
        for Action in Actions:
            layout.operator("mocapmadness.playmocap",text=Action.get("name")).action=Action.get("name")
            
class BranchMenu(bpy.types.Menu):
    bl_idname = "mocapmadness.branchmenu"
    bl_label = "Branch to Action"

    def draw(self, context):
        # Ok make the query
        action = context.object.animation_data.action
        frame_current = context.scene.frame_current
        #self.bl_label = "GGGGG"
        branches = mocapmadness.data.findall('Actions/Action[@name="%s"]/branch[@from="%.1f"]'%(action.name,frame_current))
        
        layout = self.layout
        i = 0
        for branch in branches:

                
            text = "Branch to %s frame %s"%(branch.get("action"),branch.get("to"))  
            layout.operator("mocapmadness.cycle", text=text).cyclenode=i

            i += 1   
            
class CycleMenu(bpy.types.Menu):
    bl_idname = "mocapmadness.cyclemenu"
    bl_label = "Choose a Cycle"

    def draw(self, context):
        frame_current = context.scene.frame_current        
        action = context.object.animation_data.action
        cycles = mocapmadness.data.findall('Actions/Action[@name="%s"]/cycle[@from="%.1f"]'%(action.name,frame_current))        
        layout = self.layout
        i = 0
        for cycle in cycles:
            
            text = "Cycle To %s (%s)"%(cycle.get("to"),cycle.get("match"))  
            layout.operator("mocapmadness.cycle", text=text).cyclenode=i
            #layout.operator("object.select_random", text="Random")
            i += 1
         
 
class MocapMadness_Cycle(bpy.types.Operator):
    bl_idname = "mocapmadness.cycle"
    bl_label = "Cycle"
    mocapmadness_name = bl_idname
    cyclenode = IntProperty("cyclenode",default=-1)
    next = BoolProperty("next",default=False)
    if mocapmadness.settings :
        print("SETTINGS RIPE TO PICK")
        #mocapmadness.SetOperatorProperties(MocapMadness_Cycle) 
        for cmd in mocapmadness.SetOperatorProperties("mocapmadness.cycle"):
            exec(cmd)
            
    def invoke(self, context, event):
        mocapmadness.SetOperatorSettings(self)        
        wm = context.window_manager
        #return wm.invoke_props_dialog(self)        
        if self.cyclenode > -1:
            
            return wm.invoke_props_dialog(self)
        else:
            print("clean")
            return(self.execute(context))
            #return wm.invoke_popup(self)
        return {'RUNNING_MODAL'}

    def draw(self, context):
        print("CYCLE DRAW")
        layout = self.layout
        layout.prop(self,"cyclenode")
        layout.prop(self,"Repeat")
            
    def execute(self, context):
        strip = getCurrentActionFromStrips(context,"BvhTrack")[1] 
        striptime = strip.strip_time
        print(self.cyclenode)
        if self.next :

            frame_current = context.scene.frame_current
            
            action = strip.action
            # goto the frame that has the next cycle
            cycles = mocapmadness.data.findall('Actions/Action[@name="%s"]/cycle'%(action.name))
             
            
            for cycle in cycles:
               
                goto = frame_current + float(cycle.get("to")) - striptime
                
                if goto > frame_current :
                    context.scene.frame_set(goto)
                    return {'FINISHED'}
            return {'FINISHED'} 
        
           
        if self.cyclenode > -1 :
            action = strip.action
            cycles = mocapmadness.data.findall('Actions/Action[@name="%s"]/cycle[@from="%.1f"]'%(action.name,striptime)) 
            cyclenode = cycles[self.cyclenode]
            createcycle(strip,cyclenode,self.Repeat)
                   
        return{'FINISHED'}
        #print(self.tolerance)
        #posecompare(tolerance,MIN_CYCLE_LENGTH=15,FRAMESTEP=2)
        


class MocapMadness_Branch(bpy.types.Operator):
    bl_idname = "mocapmadness.branch"
    bl_label = "branch"
    mocapmadness_name = bl_idname
    branchnode = IntProperty("branchnode",default=-1)
    next = BoolProperty("next",default=False)

    if mocapmadness.settings :
        #mocapmadness.SetOperatorProperties(MocapMadness_Branch) 
        for cmd in mocapmadness.SetOperatorProperties("mocapmadness.branch"):
            exec(cmd)
            
    def invoke(self, context, event):
        mocapmadness.SetOperatorSettings(self)        
        wm = context.window_manager
        #return wm.invoke_props_dialog(self)        
        if self.branchnode > -1:
            
            return wm.invoke_props_dialog(self)
        else:
            print("clean")
            return(self.execute(context))
            #return wm.invoke_popup(self)
        return {'RUNNING_MODAL'}

    def draw(self, context):
        #print("BRANCH DRAW")
        layout = self.layout
        layout.prop(self,"branchnode")
        layout.prop(self,"Repeat")
        layout.prop(self,"FixGimble")
        
        layout.prop(self,"FixGimbleAngle")
            
    def execute(self, context):
        strip = getCurrentActionFromStrips(context,"BvhTrack")[1] 
        striptime = strip.strip_time

        if self.next :

            frame_current = context.scene.frame_current
            
            action = strip.action
            # goto the frame that has the next cycle
            branches = mocapmadness.data.findall('Actions/Action[@name="%s"]/branch'%(action.name))
             
            
            for branch in branches:
               
                goto = frame_current + float(branch.get("to")) - striptime
                
                if goto > frame_current :
                    context.scene.frame_set(goto)
                    
            return {'FINISHED'} 
        
           
        if self.branchnode > -1 :
            action = strip.action
            branches = mocapmadness.data.findall('Actions/Action[@name="%s"]/branch[@from="%.1f"]'%(action.name,striptime)) 
            branchnode = branches[self.branchnode]
            createbranch(strip,branchnode,FixGimble=self.FixGimble,FixGimbleAngle=self.FixGimbleAngle)
                   
        return{'FINISHED'}
        

        
def register():
    bpy.utils.register_class(ActionMenu)
    bpy.utils.register_class(CycleMenu)
    bpy.utils.register_class(BranchMenu)
    bpy.utils.register_class(NLAPlayMocap)
    bpy.utils.register_class(CyclePanel)
    bpy.utils.register_class(BranchPanel)
    bpy.utils.register_class(MocapMadness_Cycle)
    bpy.utils.register_class(MocapMadness_Branch)
    bpy.utils.register_class(PlayMocapList)
    
def unregister():
    
    bpy.utils.unregister_class(ActionMenu)
    bpy.utils.unregister_class(CycleMenu)
    bpy.utils.unregister_class(BranchMenu)
    bpy.utils.unregister_class(NLAPlayMocap)
    bpy.utils.unregister_class(CyclePanel)
    bpy.utils.unregister_class(BranchPanel)
    bpy.utils.unregister_class(MocapMadness_Cycle) 
    bpy.utils.unregister_class(MocapMadness_Branch)
    bpy.utils.unregister_class(PlayMocapList)